import numpy as np
from matplotlib import pyplot as plt
from PIL import Image
import sys

sys.path.insert(0, '../../algorithms/distance/')

from rms_diff import rmsDiffPil, rmsDiffNumpy

sys.path.insert(0, '../../algorithms/context_adaptive_interpolator/')

from context_adaptive_interpolator import contextAdaptiveInterpolator
from tqdm import tqdm

IMAGE_SIZE = 64
NUMBER_OF_PHONES = 1#0
NUMBER_OF_IMAGES_PER_PHONE = 10_000
# Compared to images being 1.
PRNU_FACTOR = 0.1

# Generate PRNUs and images of phones.
# Is such `np.maximum` probabilistically correct with our theoretical method? See #19.
def randomImage(scale):
    return np.random.normal(loc = 0, scale = scale, size = (IMAGE_SIZE, IMAGE_SIZE))

imagesWithoutPrnu = [[randomImage(scale = 1) for _ in range(NUMBER_OF_IMAGES_PER_PHONE)] for phoneIndex in range(NUMBER_OF_PHONES)]

prnus = [np.array(Image.open('prnu.png').convert('F')) * PRNU_FACTOR / 255]

imagesWithPrnu = [[imageWithoutPrnu + prnus[phoneIndex] for imageWithoutPrnu in imagesWithoutPrnu[phoneIndex]] for phoneIndex in range(NUMBER_OF_PHONES)]

allImages = np.max([np.max(imagesWithoutPrnu) + np.max(prnus) + np.max(imagesWithPrnu)])

def toPilImage(npArray):
    return Image.fromarray(npArray)

def showImageWithPil(npArray):
    npArray -= npArray.min()
    npArray = (npArray / npArray.max()) * 255
    Image.fromarray(npArray).show()

def showImageWithMatplotlib(npArray):
    plt.imshow(npArray)
    plt.show()

NUMBER_OF_ROWS = 5
NUMBER_OF_COLUMNS = 3
fig, axes = plt.subplots(NUMBER_OF_ROWS, NUMBER_OF_COLUMNS)
fig.suptitle('Single PRNU estimation with images being Gaussian noise')

prnusPil = [toPilImage(prnu) for prnu in prnus]
MAIN_AXIS_ROW_INDEX = 1
mainAxis = axes[MAIN_AXIS_ROW_INDEX]
mainAxis[0].set_title('Actual PRNU')
mainAxis[0].imshow(prnus[0])

mainAxis[1].set_title('First image without PRNU')
mainAxis[1].imshow(imagesWithoutPrnu[0][0])

assert NUMBER_OF_IMAGES_PER_PHONE >= 10 ** (NUMBER_OF_ROWS - 1), 'Try to use more images than generated!'

for rowIndex, numberOfImages in enumerate([10 ** power for power in range(NUMBER_OF_ROWS)]):
    imagesWithPrnuPil0Mean = np.array(imagesWithPrnu[0][:numberOfImages]).mean(axis = 0)
    title = (f'Mean of first $N$ images with PRNU\ni.e. estimated PRNU\nRMS with actual PRNU\n\n' if rowIndex == 0 else '') + f'$N$ = {numberOfImages:,}, $RMS$ = {round(rmsDiffNumpy(imagesWithPrnuPil0Mean, prnus[0]), 4)}'
    axes[rowIndex][2].set_title(title)
    axes[rowIndex][2].imshow(imagesWithPrnuPil0Mean)

for columnIndex in range(NUMBER_OF_COLUMNS):
    for axisIndex in range(NUMBER_OF_ROWS):
        if axisIndex != MAIN_AXIS_ROW_INDEX:
            axes[axisIndex][columnIndex].axis('off')

plt.tight_layout()
plt.show()

##

def toFileName(title):
    return title.lower().replace(' ', '_').replace(',', '_')

for title, image in zip(['Actual PRNU', 'First image without PRNU'], [prnus[0], imagesWithoutPrnu[0][0]]):
    plt.title(title)
    plt.imshow(image)
    plt.savefig(title.lower().replace(' ', '_') + '.svg')

for numberOfImages in [10 ** power for power in range(NUMBER_OF_ROWS)]:
    title = 'First image with PRNU' if numberOfImages == 1 else f'Mean of first {numberOfImages:,} images with PRNU'
    image = np.array(imagesWithPrnu[0][:numberOfImages]).mean(axis = 0)
    plt.title(f'{title}\ni.e. estimated PRNU\nRMS with actual PRNU = {round(rmsDiffNumpy(image, prnus[0]), 4)}')
    plt.imshow(image)
    plt.savefig(f'{toFileName(title)}.svg')

##

# Compute CAI of phone images.
caiImages = [[contextAdaptiveInterpolator(image.load(), image) for image in imagesWithPrnuPil[phoneIndex]] for phoneIndex in tqdm(range(NUMBER_OF_PHONES))]
#caiImages[0][0].show()

# Guess the phone by consider the one reaching the lowest RMS difference between the estimated PRNU and the actual one.
def getPhoneIndexByNearestPrnu(estimatedPrnu):
    nearestPrnuRmsDiff = None
    nearestPrnuPhoneIndex = None
    for phoneIndex, prnu in enumerate(prnusPil):
        prnuRmsDiff = rmsDiffPil(estimatedPrnu, prnu)
        if nearestPrnuRmsDiff is None or prnuRmsDiff < nearestPrnuRmsDiff:
            nearestPrnuRmsDiff = prnuRmsDiff
            nearestPrnuPhoneIndex = phoneIndex
    return nearestPrnuPhoneIndex

# Predict the phones based on the estimated PRNUs and compute each method accuracy.
# What about CAI on `phoneImagesMean`? See https://gitea.lemnoslife.com/Benjamin_Loison/Robust_image_source_identification_on_modern_smartphones/issues/9#issuecomment-1369.
correctGuessesByMeanImages = 0
correctGuessesByMeanCAIImages = 0
# Maybe make sense only because Gaussian images here.
correctGuessesByCAIImagesMean = 0
for phoneIndex in range(NUMBER_OF_PHONES):
    phoneImages = imagesWithPrnuPil[phoneIndex]
    phoneCaiImages = caiImages[phoneIndex]

    phoneImagesMean = toPilImage(np.array(phoneImages).mean(axis = 0))
    caiImagesMean = toPilImage(np.array(phoneCaiImages).mean(axis = 0))
    caiOverPhoneImagesMean = contextAdaptiveInterpolator(phoneImagesMean.load(), phoneImagesMean)

    phonePrnu = prnusPil[phoneIndex]

    print('RMS diff with mean image =', rmsDiffPil(phoneImagesMean, phonePrnu))
    print('RMS diff with mean CAI images =', rmsDiffPil(caiImagesMean, phonePrnu))
    print('RMS diff with CAI images mean =', rmsDiffPil(caiOverPhoneImagesMean, phonePrnu))

    guessedPhoneIndexByMeanImages = getPhoneIndexByNearestPrnu(phoneImagesMean)
    guessedPhoneIndexByMeanCAIImages = getPhoneIndexByNearestPrnu(caiImagesMean)
    guessedPhoneIndexByCAIImagesMean = getPhoneIndexByNearestPrnu(caiOverPhoneImagesMean)

    print(f'Actual phone index {phoneIndex}, guessed phone index {guessedPhoneIndexByMeanImages} by mean images, {guessedPhoneIndexByMeanCAIImages} by mean CAI images and {guessedPhoneIndexByCAIImagesMean} by CAI images mean')

    correctGuessesByMeanImages += 1 if phoneIndex == guessedPhoneIndexByMeanImages else 0
    correctGuessesByMeanCAIImages += 1 if phoneIndex == guessedPhoneIndexByMeanCAIImages else 0
    correctGuessesByCAIImagesMean += 1 if phoneIndex == guessedPhoneIndexByCAIImagesMean else 0

print(f'{correctGuessesByMeanImages / NUMBER_OF_PHONES=}')
print(f'{correctGuessesByMeanCAIImages / NUMBER_OF_PHONES=}')
print(f'{correctGuessesByCAIImagesMean / NUMBER_OF_PHONES=}')

